# visuals.py

import numpy as np
from OpenGL.GL import *


class Circle:
    def __init__(self, center, radius, fill=True, thickness=1.0, color=(255.0, 255.0, 255.0), num_segments=100):
        """
        Initialize a Circle object.

        Parameters:
            center: Center coordinates of the circle as a tuple (x, y).
            radius: Radius of the circle.
            fill: If True, the circle will be filled. If False, only the outline will be drawn.
            thickness: Thickness of the outline if fill is False.
            color: RGB tuple (r, g, b) for the circle's color, with values between 0 and 255.
            num_segments: The number of line segments to approximate the circle.
        """
        self.center = np.asarray(center, dtype=np.float32)
        self.radius = float(radius)
        self.fill = fill
        self.thickness = thickness
        self.set_color(color)
        self.num_segments = int(num_segments)

    def set_center(self, center):
        self.center = np.asarray(center, dtype=np.float32)

    def set_radius(self, radius):
        self.radius = float(radius)

    def set_color(self, color):
        self.color = np.asarray(color) / 255.0  # Normalize color to [0, 1]

    def draw(self):
        
        # Set the texture environment mode to GL_MODULATE
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE)

        # Set the color for the circle
        glColor3f(*self.color)

        # Choose the OpenGL primitive type based on whether the circle is filled or not
        if self.fill:
            glBegin(GL_TRIANGLE_FAN)
            # Start with the center point for a filled circle
            glVertex2f(*self.center)
        else:
            glLineWidth(self.thickness)
            glBegin(GL_LINE_LOOP)

        # Draw the circle by specifying vertices
        for i in range(self.num_segments + 1):
            angle = 2.0 * np.pi * i / self.num_segments  # Angle for each segment
            x = self.center[0] + self.radius * np.cos(angle)
            y = self.center[1] + self.radius * np.sin(angle)
            glVertex2f(x, y)

        glEnd()


class Rectangle:
    def __init__(self, a_rect, fill=True, thickness=1.0, color=(255.0, 255.0, 255.0)):
        """
        Initialize a Rectangle object.

        Parameters:
            a_rect: A rectangle defined as [x1, y1, x2, y2] or [[x1, y1], [x2, y2]].
            fill: If True, the rectangle will be filled. If False, only the outline will be drawn.
            thickness: Thickness of the outline if fill is False.
            color: RGB tuple (r, g, b) for the rectangle's color, with values between 0 and 255.
        """
        self.set_rect(a_rect)
        self.fill = fill
        self.thickness = thickness
        self.set_color(color)

    def set_rect(self, a_rect):
        a_rect = np.asarray(a_rect)
        if a_rect.shape[0] == 4:
            self.x1, self.y1, self.x2, self.y2 = a_rect
        elif a_rect.shape == (2, 2):
            (self.x1, self.y1), (self.x2, self.y2) = a_rect
        else:
            raise ValueError("Rectangle must be [x1, y1, x2, y2] or [[x1, y1], [x2, y2]].")

        if self.x2 < self.x1 or self.y2 < self.y1:
            raise ValueError("x2 must be >= x1 and y2 must be >= y1.")

    def set_color(self, color):
        self.color = np.asarray(color) / 255.0

    def draw(self):

        # Set the texture environment mode to GL_MODULATE
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE)
        glColor3f(*self.color)

        if self.fill:
            glBegin(GL_QUADS)
            glVertex2f(self.x1, self.y1)  # Top-left
            glVertex2f(self.x2, self.y1)  # Top-right
            glVertex2f(self.x2, self.y2)  # Bottom-right
            glVertex2f(self.x1, self.y2)  # Bottom-left
            glEnd()
        else:
            glLineWidth(self.thickness)
            glBegin(GL_LINE_LOOP)
            glVertex2f(self.x1, self.y1)  # Top-left
            glVertex2f(self.x2, self.y1)  # Top-right
            glVertex2f(self.x2, self.y2)  # Bottom-right
            glVertex2f(self.x1, self.y2)  # Bottom-left
            glEnd()
        


class Line:
    def __init__(self, start_point, end_point, thickness=1.0, color=(255.0, 0.0, 0.0)):
        """
        Initialize a Line object.

        Parameters:
            start_point: Starting point of the line as a tuple (x, y).
            end_point: Ending point of the line as a tuple (x, y).
            thickness: Thickness of the line.
            color: RGB tuple (r, g, b) for the line color, with values between 0 and 255.
        """
        self.start_point = np.asarray(start_point, dtype=np.float32)
        self.end_point = np.asarray(end_point, dtype=np.float32)
        self.thickness = thickness
        self.set_color(color)

    def set_start_point(self, point):
        self.start_point = np.asarray(point, dtype=np.float32)

    def set_end_point(self, point):
        self.end_point = np.asarray(point, dtype=np.float32)

    def set_color(self, color):
        self.color = np.asarray(color) / 255.0  # Normalize color to [0, 1]

    def set_thickness(self, thickness):
        self.thickness = thickness

    def draw(self):
        
        # Set the texture environment mode to GL_MODULATE
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE)
        glColor3f(*self.color)
        glLineWidth(self.thickness)
        glBegin(GL_LINES)
        glVertex2f(*self.start_point)
        glVertex2f(*self.end_point)
        glEnd()



class FixationCross:
    def __init__(self, center, half_width, half_height, thickness=1.0, color=(255.0, 0.0, 0.0)):
        """
        Initialize a FixationCross object.

        Parameters:
            center: Center point of the fixation cross as a tuple (x, y).
            half_width: Half-width of the fixation cross.
            half_height: Half-height of the fixation cross.
            thickness: Thickness of the cross lines.
            color: RGB tuple (r, g, b) for the cross color, with values between 0 and 255.
        """
        self.center = np.asarray(center, dtype=np.float32)
        self.half_width = half_width
        self.half_height = half_height
        self.thickness = thickness
        self.color = color

        # Create the horizontal and vertical lines
        self.horizontal_line = Line(
            start_point=(self.center[0] - self.half_width, self.center[1]),
            end_point=(self.center[0] + self.half_width, self.center[1]),
            thickness=self.thickness,
            color=self.color
        )

        self.vertical_line = Line(
            start_point=(self.center[0], self.center[1] - self.half_height),
            end_point=(self.center[0], self.center[1] + self.half_height),
            thickness=self.thickness,
            color=self.color
        )

    def set_center(self, center):
        self.center = np.asarray(center, dtype=np.float32)
        self.update_lines()

    def set_size(self, half_width, half_height):
        self.half_width = half_width
        self.half_height = half_height
        self.update_lines()

    def set_color(self, color):
        self.color = color
        self.horizontal_line.set_color(color)
        self.vertical_line.set_color(color)

    def set_thickness(self, thickness):
        self.thickness = thickness
        self.horizontal_line.set_thickness(thickness)
        self.vertical_line.set_thickness(thickness)

    def update_lines(self):
        # Update the positions of the lines based on the center and size
        self.horizontal_line.set_start_point((self.center[0] - self.half_width, self.center[1]))
        self.horizontal_line.set_end_point((self.center[0] + self.half_width, self.center[1]))
        self.vertical_line.set_start_point((self.center[0], self.center[1] - self.half_height))
        self.vertical_line.set_end_point((self.center[0], self.center[1] + self.half_height))

    def draw(self):
        # Draw both lines
        self.horizontal_line.draw()
        self.vertical_line.draw()

def center_rect_on_point(a_rect, a_point):
    """
    Center a rectangle on a point.

    Parameters:
        a_rect: A rectangle defined as [x1, y1, x2, y2] or [[x1, y1], [x2, y2]].
        a_point: A point defined as [x, y].
    """
    x1, y1, x2, y2 = a_rect if len(a_rect) == 4 else [a_rect[0][0], a_rect[0][1], a_rect[1][0], a_rect[1][1]]
    width = x2 - x1
    height = y2 - y1
    cx, cy = a_point

    new_x1 = cx - width // 2
    new_y1 = cy - height // 2
    new_x2 = new_x1 + width
    new_y2 = new_y1 + height

    return [new_x1, new_y1, new_x2, new_y2]



# __init__.py

from .screen import Screen
from .textures import Texture
from .shapes import (
    Circle,
    Rectangle,
    Line,
    FixationCross,
    center_rect_on_point,
)
from .responses import ResponseHandler, Scrollbar
from .text import Text, TextBox
from .psychophysics import (
    fabriquer_grille_sin,
    fabriquer_gabor,
    noisy_bit_dithering,
    fabriquer_enveloppe_gaussienne,
    stretch,
    fabriquer_wiggles_sin,
    fabriquer_cercles_sin,
    fabriquer_secteurs_sin,
    fabriquer_grand_damier,
    fabriquer_petit_damier,
    location_bubbles,
    )
from .audio import Audio
from .quest import QuestObject



# textures.py
import numpy as np
from OpenGL.GL import *


class Texture:
    def __init__(self, image):
        self.texture_id = glGenTextures(1)
        self.load_texture(image)
    
    def load_texture(self, image):
        glBindTexture(GL_TEXTURE_2D, self.texture_id)

        glPixelStorei(GL_UNPACK_ALIGNMENT, 1)

        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, image.shape[1], image.shape[0], 0, GL_RGB, GL_UNSIGNED_BYTE, image)
        
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
        glBindTexture(GL_TEXTURE_2D, 0)

    def bind(self):
        glBindTexture(GL_TEXTURE_2D, self.texture_id)
        glEnable(GL_TEXTURE_2D)

    def unbind(self):
        glDisable(GL_TEXTURE_2D)
        glBindTexture(GL_TEXTURE_2D, 0)
        # Reset the modelview matrix to avoid any transformations being carried over
        glLoadIdentity()

    def delete(self):
        glDeleteTextures([self.texture_id])

    def draw(self, a_rect):
        """
        Draw a texture on the screen.
        Parameters:
            texture_id: The ID of the texture to draw.
            a_rect: A rectangle defined as [x1, y1, x2, y2] or [[x1, xy], [x2, y2]].
        """

        a_rect = np.asarray(a_rect)
        if a_rect.shape[0]==4:
            x1, y1, x2, y2 = np.asarray(a_rect)
        elif a_rect.shape[0]==2 & a_rect.shape[1]==2:
            x1, y1 = a_rect[0]
            x2, y2 = a_rect[1]
        else:
            raise ValueError("A rectangle is defined either as [x1, y1, x2, y2] or [[x1, xy], [x2, y2]].")

        if x2<=x1 or y2<=y1:
            raise ValueError("x2 must be equal or smaller than x1 an y2 must be equal or smaller than y1.")

        # Enable texturing modulations        
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE)

        # Open all colour channels.
        glColor3f(1.0, 1.0, 1.0)

        # bind the texture to be drawn       
        self.bind()
        
        # map the texture to the rectangle
        glBegin(GL_QUADS)
        # Compute centered vertex positions
        glTexCoord2f(0, 0); glVertex2f(x1, y1) # left - top
        glTexCoord2f(1, 0); glVertex2f(x2, y1) # right - top
        glTexCoord2f(1, 1); glVertex2f(x2, y2) # right - bottom
        glTexCoord2f(0, 1); glVertex2f(x1, y2) # left - bottom
        glEnd()
        
        # unbind the texture
        self.unbind()





# screen.py

import os
import pygame
from pygame.locals import *
from OpenGL.GL import *
from OpenGL.GLU import *
import numpy as np
from screeninfo import get_monitors
from time import monotonic_ns, sleep


class Screen:
    def __init__(self, screen_number=0, width=None, height=None, fullscreen=True, vsync=True, desired_refresh_rate=60):
        """
        Initialize a screen using Pygame and OpenGL.
        
        Parameters
        ----------
        
        screen_number : int
            The number of the screen to use. Default is 0.
            
        width : int
            The width of the screen. Default is the width of the monitor.
        
        height : int
            The height of the screen. Default is the height of the monitor.

        fullscreen : bool
            Whether to use fullscreen mode. Default is True.
        
        vsync : bool
            Whether to use vertical synchronization. Default is True.
        
        desired_refresh_rate : int
            The desired refresh rate of the screen in Hz. Default is 60.
        
        """
        # Get monitors
        monitors = get_monitors()
        
        # Set monitor
        if len(monitors) > screen_number:
            monitor = monitors[screen_number]
        else:
            monitor = monitors[0]
        
        self.monitor = monitor
        self.width = width or monitor.width
        self.height = height or monitor.height
        self.fullscreen = fullscreen
        self.vsync = vsync
        self.desired_refresh_rate = desired_refresh_rate
        self.mouse_visible = True

        # Internal timing variables for frame measurement
        self.last_flip_time = None
        self.prev_flip_time = None

        # Set the window position to the monitor's position
        os.environ['SDL_VIDEO_WINDOW_POS'] = f"{monitor.x},{monitor.y}"

        # Initialize Pygame and create a window
        pygame.init()
        flags = DOUBLEBUF | OPENGL
        if fullscreen:
            flags |= FULLSCREEN
        self.screen = pygame.display.set_mode((self.width, self.height), flags, vsync=vsync)
        self.clock = pygame.time.Clock()

        # grab the mouse attention
        pygame.event.set_grab(True)

        # Initialize OpenGL
        glViewport(0, 0, self.width, self.height)  # Set the viewport to match the screen dimensions


        # Initialize OpenGL
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluOrtho2D(0, self.width, self.height, 0)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        glClearColor(0.5, 0.5, 0.5, 1)  # Gray background

        # Clear the color buffer initially (sets the background color for the first frame)
        glClear(GL_COLOR_BUFFER_BIT)

        # Disable depth testing since it's not needed for 2D rendering
        glDisable(GL_DEPTH_TEST)

        # Disable textures initially to avoid unexpected textures appearing
        glDisable(GL_TEXTURE_2D)

        # Enable blending for transparency handling (optional, but useful for 2D graphics)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

    def flip(self):
        """
        Flip the screen and keep a timestamp.

        Returns
        -------
        float
            The timestamp of the flip.
        """
        pygame.display.flip()
        self.tick()
        self.prev_flip_time = self.last_flip_time
        this_time = monotonic_ns()
        self.last_flip_time = this_time
        return this_time

    def get_flip_interval(self):
        """
        Get the interval between the last two flips.

        Returns
        -------
        float
            The interval between the last two flips in seconds.
        
        """
        if self.last_flip_time is None or self.prev_flip_time is None:
            return None
        return (self.last_flip_time - self.prev_flip_time)/1e9 # go back to seconds
    
    def fill(self, color=(128,128,128)):
        """
        Fill the screen with a color.

        Parameters
        ----------
        color : tuple
            The color to fill the screen with. Should be a tuple of 3 integers between 0 and 255.

        """
        # Use OpenGL to clear the screen with the specified color
        glBindTexture(GL_TEXTURE_2D, 0)
        glDisable(GL_TEXTURE_2D)
        # glDisable(GL_BLEND)
        # glBlendFunc(GL_ONE, GL_ZERO)
        glDisable(GL_DEPTH_TEST)
        glClearColor(color[0]/255.0, color[1]/255.0, color[2]/255.0, 1.0)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

    def tick(self):
        """
        Limit the frame rate to the desired refresh rate
        """
        self.clock.tick(self.desired_refresh_rate)

    def test_flip_intervals(self, num_frames=50):
        """
        Test the flip intervals of the screen.

        Parameters
        ----------
        num_frames : int
            The number of frames to test. Default is 50.
        """
        frame_rates = []
        for _ in range(num_frames):
            self.fill((128, 128, 128))
            self.flip()
            frame_rates.append(self.get_flip_interval())
        frame_rates_array = np.array(frame_rates)
        frame_rate_actual = np.mean(frame_rates_array[frame_rates_array>0])
        return frame_rate_actual

    def hide_mouse(self):
        # Hide the mouse cursor
        pygame.mouse.set_visible(False)

    def show_mouse(self):
        # Show the mouse cursor
        pygame.mouse.set_visible(True)

    def wait(self, duration_secs):
        """
        Wait for a specified duration in seconds using high-precision timing.
        
        Parameters:
            duration_secs: The duration to wait in seconds (float).
        """
        start_time = monotonic_ns()
        end_time = start_time + int(duration_secs * 1e9)  # Convert seconds to nanoseconds

        # Sleep in small increments to reduce CPU usage
        while True:
            current_time = monotonic_ns()
            remaining_time = (end_time - current_time) / 1e9  # Convert to seconds

            if remaining_time <= 0:
                break

            if remaining_time > 0.005:
                # If more than 5 milliseconds remaining, sleep for a short duration
                sleep(0.001)  # Sleep for 1 millisecond
            else:
                # Busy-wait for the remaining time for higher precision
                pass

    def close(self):
        """
        Close the screen.
        """
        self.show_mouse()
        pygame.quit()


import numpy as np
import pygame
from OpenGL.GL import *
from OpenGL.GLU import *


class Text:
    def __init__(self, text, font_name='Helvetica', font_size=24, color=(255, 255, 255), dest_rect=None, line_spacing=4):
        """
        Initialize the Text object.

        Parameters:
            text: 
                The text string to render.
            dest_rect : list or tuple
                A rectangle [x1, y1, x2, y2] defining the text bounding box. Optional.
            font_name: 
                the name of the font.
            font_size:
                The font size in points.
            color: 
                Color of the text as an RGB tuple.
            
        """
        pygame.font.init()
        self.text = text
        self.font_name = font_name
        self.font_size = font_size
        self.color = color
        self.dest_rect = dest_rect
        self.texture_id = None
        self.surface = None
        self.line_spacing = line_spacing

        self.lines = []
        self._split_text_into_lines()
        self._generate_surface()

    def _split_text_into_lines(self):
        """Split the text into multiple lines that fit within the dest_rect width."""
        if not self.dest_rect:
            raise ValueError("dest_rect must be set to split text into lines.")

        font = pygame.font.SysFont(self.font_name, self.font_size)
        max_width = self.dest_rect[2] - self.dest_rect[0]  # Width of the rectangle

        words = self.text.split()
        line = ""
        self.lines = []
        for word in words:
            test_line = f"{line} {word}".strip()
            if font.size(test_line)[0] <= max_width:
                line = test_line
            else:
                self.lines.append(line)
                line = word
        if line:
            self.lines.append(line)

    def _generate_surface(self):
        """Generate a Pygame surface with the text rendered."""
        font = pygame.font.SysFont(self.font_name, self.font_size)

        line_surfaces = [font.render(line, True, self.color) for line in self.lines]
        max_width = max(surface.get_width() for surface in line_surfaces)
        total_height = sum(surface.get_height() for surface in line_surfaces) + (len(line_surfaces) - 1) * self.line_spacing

        # Create the surface with an appropriate size
        self.surface = pygame.Surface((max_width, total_height), pygame.SRCALPHA)
        self.surface.fill((0, 0, 0, 0))  # Transparent background

        # Draw each line on the surface
        y_offset = 0
        for surface in line_surfaces:
            self.surface.blit(surface, (0, y_offset))
            y_offset += surface.get_height() + self.line_spacing

        self._generate_texture()

    def _generate_texture(self):
        """Upload the Pygame surface as an OpenGL texture."""
        if self.texture_id:
            glDeleteTextures([self.texture_id])

        self.texture_id = glGenTextures(1)
        glBindTexture(GL_TEXTURE_2D, self.texture_id)

        surface_data = pygame.image.tostring(self.surface, "RGBA", True)
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, self.surface.get_width(),
                     self.surface.get_height(), 0, GL_RGBA, GL_UNSIGNED_BYTE, surface_data)

        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)

        glBindTexture(GL_TEXTURE_2D, 0)

    def draw(self):
        """
        Draw the text on the screen, centered within the dest_rect.
        """
        if not self.dest_rect:
            raise ValueError("dest_rect must be set to draw text.")

        x1, y1, x2, y2 = self.dest_rect
        rect_width = x2 - x1
        rect_height = y2 - y1

        # Center the text within the rectangle
        texture_width = self.surface.get_width()
        texture_height = self.surface.get_height()
        center_x = x1 + rect_width // 2
        center_y = y1 + rect_height // 2
        x_start = center_x - texture_width // 2
        y_start = center_y - texture_height // 2

        # Enable blending and set blend function for transparency
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        # Bind the texture and draw the quad
        glEnable(GL_TEXTURE_2D)
        glBindTexture(GL_TEXTURE_2D, self.texture_id)

        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE)

        glBegin(GL_QUADS)
        glTexCoord2f(0, 1); glVertex2f(x_start, y_start)  # Top-left
        glTexCoord2f(1, 1); glVertex2f(x_start + texture_width, y_start)  # Top-right
        glTexCoord2f(1, 0); glVertex2f(x_start + texture_width, y_start + texture_height)  # Bottom-right
        glTexCoord2f(0, 0); glVertex2f(x_start, y_start + texture_height)  # Bottom-left
        glEnd()

        # Unbind texture and disable states
        glBindTexture(GL_TEXTURE_2D, 0)
        glDisable(GL_TEXTURE_2D)
        glDisable(GL_BLEND)

    def set_text(self, new_text):
        """
        Update the text and regenerate the surface and texture.
        """
        self.text = new_text
        self._split_text_into_lines()
        self._generate_surface()

    def set_dest_rect(self, dest_rect):
        """
        Update the destination rectangle and regenerate the surface if needed.
        """
        self.dest_rect = dest_rect
        self._split_text_into_lines()
        self._generate_surface()
        



class TextBox:
    """
    A class to create an interactive text input box for user input.

    Attributes:
        position (tuple): The (x, y) position of the top-left corner of the text box.
        size (tuple): The (width, height) dimensions of the text box.
        font (pygame.font.Font): The font used for rendering text.
        text_color (tuple): RGB color of the input text.
        box_color (tuple): RGB color of the text box background.
        border_color (tuple): RGB color of the text box border.
        border_thickness (int): Thickness of the border in pixels.
        text (str): The current text input by the user.
        submitted (bool): Flag indicating whether the user has submitted the text.
        cursor_position (int): The position of the cursor within the text.
    """

    def __init__(self, position, size, font, text_color=(0, 0, 0),
                 box_color=(255, 255, 255), border_color=(0, 0, 0), border_thickness=2):
        """
        Initialize the TextBox object.

        Parameters:
            position (tuple): The (x, y) position of the top-left corner of the text box.
            size (tuple): The (width, height) dimensions of the text box.
            font (pygame.font.Font): The font used for rendering text.
            text_color (tuple): RGB color of the input text.
            box_color (tuple): RGB color of the text box background.
            border_color (tuple): RGB color of the text box border.
            border_thickness (int): Thickness of the border in pixels.
        """
        self.position = position
        self.size = size
        self.font = font
        self.text_color = text_color
        self.box_color = box_color
        self.border_color = border_color
        self.border_thickness = border_thickness

        self.text = ""  # Current text input by the user
        self.text_surface = None  # Surface for rendered text
        self.texture_id = None  # OpenGL texture ID
        self.width = self.size[0]  # Width of the texture (same as text box width)
        self.height = self.size[1]  # Height of the texture (same as text box height)
        self.cursor_visible = True  # Cursor visibility flag
        self.cursor_timer = 0  # Timer for cursor blinking
        self.cursor_switch_ms = 500  # Cursor blink interval in milliseconds
        self.submitted = False  # Flag to indicate submission

        self.padding = 5  # Padding inside the text box
        self.offset = 0  # Offset for scrolling text

        self.cursor_position = 0  # Cursor position within the text

        self.update_texture()  # Initial texture update

    def handle_event(self, event):
        """
        Handle Pygame events related to text input.

        Parameters:
            event (pygame.event.Event): A Pygame event object.
        """
        if event.type == pygame.KEYDOWN:
            if event.key == pygame.K_BACKSPACE:
                if self.cursor_position > 0:
                    # Remove character before the cursor
                    self.text = self.text[:self.cursor_position - 1] + self.text[self.cursor_position:]
                    self.cursor_position -= 1
                    self.update_texture()
            elif event.key == pygame.K_DELETE:
                if self.cursor_position < len(self.text):
                    # Remove character after the cursor
                    self.text = self.text[:self.cursor_position] + self.text[self.cursor_position + 1:]
                    self.update_texture()
            elif event.key == pygame.K_RETURN or event.key == pygame.K_KP_ENTER:
                # Enter key pressed; set submitted flag to True
                self.submitted = True
            elif event.key == pygame.K_LEFT:
                if self.cursor_position > 0:
                    self.cursor_position -= 1
                    self.update_texture()
            elif event.key == pygame.K_RIGHT:
                if self.cursor_position < len(self.text):
                    self.cursor_position += 1
                    self.update_texture()
            else:
                # Add the unicode character to the text if it's printable
                char = event.unicode
                if char.isprintable():
                    self.text = self.text[:self.cursor_position] + char + self.text[self.cursor_position:]
                    self.cursor_position += len(char)
                    self.update_texture()

    def update_texture(self):
        """
        Update the OpenGL texture with the current text.
        """
        # Render the text onto a surface
        self.text_surface = self.font.render(self.text, True, self.text_color)
        text_width, text_height = self.text_surface.get_size()

        # Create a surface with the size of the text box
        box_surface = pygame.Surface(self.size, pygame.SRCALPHA)

        # Calculate maximum width available for text
        max_text_width = self.size[0] - 2 * self.padding

        # Calculate the width of text up to the cursor position
        cursor_text = self.text[:self.cursor_position]
        cursor_width = self.font.size(cursor_text)[0]

        # Determine if text needs to be scrolled
        if cursor_width - self.offset > max_text_width:
            # Move offset to the right to keep cursor visible
            self.offset = cursor_width - max_text_width + 10  # Additional padding
        elif cursor_width - self.offset < 0:
            # Move offset to the left to keep cursor visible
            self.offset = cursor_width - 10  # Additional padding

        # Ensure offset is not negative
        if self.offset < 0:
            self.offset = 0

        # Blit the text onto the box surface with the calculated offset
        box_surface.blit(self.text_surface, (-self.offset + self.padding, (self.size[1] - text_height) / 2))

        # Update the texture
        self.width, self.height = box_surface.get_size()
        # Convert the surface to a string buffer for OpenGL
        text_data = pygame.image.tostring(box_surface, "RGBA", True)
        # Generate a texture ID if not already created
        if self.texture_id is None:
            self.texture_id = glGenTextures(1)
        # Bind the texture and upload the data
        glBindTexture(GL_TEXTURE_2D, self.texture_id)
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, self.width, self.height,
                     0, GL_RGBA, GL_UNSIGNED_BYTE, text_data)
        # Set texture parameters for scaling
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        # Unbind the texture
        glBindTexture(GL_TEXTURE_2D, 0)

    def update(self, delta_time):
        """
        Update the text box, handling cursor blinking.

        Parameters:
            delta_time (float): Time elapsed since the last update in milliseconds.
        """
        # Update the cursor timer
        self.cursor_timer += delta_time
        if self.cursor_timer >= self.cursor_switch_ms:
            # Reset the timer and toggle cursor visibility
            self.cursor_timer %= self.cursor_switch_ms
            self.cursor_visible = not self.cursor_visible

    def draw(self):
        """
        Draw the text box, including the background, border, text, and cursor.
        """
        x, y = self.position
        width, height = self.size

        # Draw the text box background
        glColor3ub(*self.box_color)
        glBegin(GL_QUADS)
        glVertex2f(x, y)
        glVertex2f(x + width, y)
        glVertex2f(x + width, y + height)
        glVertex2f(x, y + height)
        glEnd()

        # Draw the border around the text box if border thickness is greater than zero
        if self.border_thickness > 0:
            glColor3ub(*self.border_color)
            glLineWidth(self.border_thickness)
            glBegin(GL_LINE_LOOP)
            glVertex2f(x, y)
            glVertex2f(x + width, y)
            glVertex2f(x + width, y + height)
            glVertex2f(x, y + height)
            glEnd()

        # Draw the text inside the text box
        if self.texture_id is not None:
            glPushAttrib(GL_ENABLE_BIT | GL_COLOR_BUFFER_BIT | GL_TEXTURE_BIT)
            glPushMatrix()
            # Enable blending for transparency
            glEnable(GL_BLEND)
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
            # Bind the texture
            glBindTexture(GL_TEXTURE_2D, self.texture_id)
            glEnable(GL_TEXTURE_2D)
            glColor3f(1.0, 1.0, 1.0)

            # Draw the textured quad with the text
            x1 = x
            y1 = y
            x2 = x + self.width
            y2 = y + self.height

            glBegin(GL_QUADS)
            glTexCoord2f(0, 1); glVertex2f(x1, y1)
            glTexCoord2f(1, 1); glVertex2f(x2, y1)
            glTexCoord2f(1, 0); glVertex2f(x2, y2)
            glTexCoord2f(0, 0); glVertex2f(x1, y2)
            glEnd()

            # Disable textures and blending
            glDisable(GL_TEXTURE_2D)
            glBindTexture(GL_TEXTURE_2D, 0)
            glDisable(GL_BLEND)
            glPopMatrix()
            glPopAttrib()

        # Draw the cursor
        if self.cursor_visible:
            self.draw_cursor()

    def draw_cursor(self):
        """
        Draw the cursor at the appropriate position.
        """
        # Calculate cursor position
        x, y = self.position
        padding = self.padding
        text_height = self.text_surface.get_height()
        cursor_height = text_height
        cursor_width = 2  # Width of the cursor line

        # Calculate the width of text up to the cursor position
        cursor_text = self.text[:self.cursor_position]
        cursor_x_in_text = self.font.size(cursor_text)[0]

        # Calculate cursor x position on the screen
        cursor_x_position = x + padding + cursor_x_in_text - self.offset

        # Ensure cursor is within the text box
        cursor_x_min = x + padding
        cursor_x_max = x + self.size[0] - padding
        if cursor_x_position < cursor_x_min:
            cursor_x_position = cursor_x_min
        elif cursor_x_position > cursor_x_max:
            cursor_x_position = cursor_x_max

        # Set cursor color (same as text color)
        glColor3ub(*self.text_color)

        # Draw the cursor as a rectangle
        glBegin(GL_QUADS)
        glVertex2f(cursor_x_position, y + (self.size[1] - cursor_height) / 2)
        glVertex2f(cursor_x_position + cursor_width, y + (self.size[1] - cursor_height) / 2)
        glVertex2f(cursor_x_position + cursor_width, y + (self.size[1] + cursor_height) / 2)
        glVertex2f(cursor_x_position, y + (self.size[1] + cursor_height) / 2)
        glEnd()

    def clear(self):
        """
        Clear the text box content and reset the submitted flag.
        """
        self.text = ""
        self.cursor_position = 0
        self.submitted = False
        self.update_texture


import os
import time
import numpy as np
from tachypy import (
    Audio,
    Screen,
    Texture,
    Circle,
    Rectangle,
    FixationCross,
    Text,
    center_rect_on_point,
    ResponseHandler,
    fabriquer_gabor,
    noisy_bit_dithering
)

# which screen should we draw to?
screen_number = 1
screen = Screen(screen_number=screen_number, fullscreen=True, desired_refresh_rate=60)

# get some relevant screen properties
center_x = screen.width//2 
center_y = screen.height//2 

# let's initialise our FixationCross
fixation_cross = FixationCross(center=[center_x, center_y], half_width=50, half_height=50, thickness=2.0, color=(255, 0, 0))  # Red cross

# let's add a white circle
circle = Circle(center=(320, 240), radius=50, fill=True, color=(255, 255, 255))  # Green circle

# let's add a nice blue rectangle
# first we can define the rectangle it will strecth to:
a_rect = [center_x+500, 190, center_x+1000, 290]
rectangle = Rectangle(a_rect, fill=True, thickness=1.0, color=(20.0, 50.0, 255.0))

# let's start our audio player
audio_player = Audio(sample_rate=44100, channels=1)

# make a sinewave for the sound
duration = 1.0  # seconds
frequency = 440.0  # Hz (A4 note)
sample_rate = 44100
t = np.linspace(0, duration, int(sample_rate * duration), endpoint=False)
amplitude = 0.5
waveform = amplitude * np.sin(2 * np.pi * frequency * t).astype(np.float32)


# set the screen's background colour to gray
screen.fill([128, 128, 128])

# flip the screen to make the background color visible
screen.flip()

# check for the screen's actual refresh rate
frame_rate_actual = 1/screen.test_flip_intervals(num_frames=100)
print(frame_rate_actual)

# Initialize ResponseHandler
response_handler = ResponseHandler()

# create a moving Gabor patch animation
nx = 750
frequency_per_im = 10
speed_in_cycles_per_s = 3
nb_frames_per_cycle = int(frame_rate_actual / speed_in_cycles_per_s)

rms_target = 0.08
film = []
for ii in range(nb_frames_per_cycle):
    phase = 2 * np.pi * ii / nb_frames_per_cycle
    gabor = fabriquer_gabor(nx, frequence=frequency_per_im, phase=phase, angle=np.pi/4, ecart_type=0.2)

    gabor = rms_target / np.std(gabor) * (gabor - 0.5) + 0.5
    gabor_dithered = noisy_bit_dithering(gabor)
    
    gabor_rgb = np.stack((gabor_dithered,)*3, axis=-1) # could be done in the OpenGL functions, like the colors
    # gabor_rgb[:,:,2] = 255 - gabor_rgb[:,:,2]
    film.append(gabor_rgb)


# Load stimuli (example: red, green, blue squares)
textures = [Texture(stimulus) for stimulus in film]

# define the position in which the Texture will be mapped.
dest_rect = center_rect_on_point([0, 0, nx-1, nx-1], [center_x, center_y])

# Main loop
running = True

# Track frame timestamps to measure interval consistency
frame_intervals = []

# flip an initial screen and set initial time
start_time = screen.flip()

while running:
    for current_trial, texture in enumerate(textures):

        screen.fill([128, 128, 128])

        # draw a circle
        circle.draw()

        # draw a rectangle
        rectangle.draw()


        # draw the texture
        texture.draw(dest_rect)

        # draw the fixation cross
        fixation_cross.draw()

        time_stamp = screen.flip()

        frame_intervals.append(screen.get_flip_interval()) 

        # Handle events
        response_handler.get_events()
        if response_handler.should_quit():
            running = False
            break

        # Example: Check if the a key was pressed
        if response_handler.is_key_down('a'):
            print("a key pressed!")
            audio_player.play(waveform)
            # Do something in response to the spacebar press
        
   #time.sleep(0.01)

# Create a Text object
dest_rect = center_rect_on_point([0, 0, .85*screen.width, .2*screen.height], [center_x, center_y])
text = Text("Thanks for using tachypy. Here is a long sentence that should be rendered on multiple lines. Only if that is necessary, of course. Press spacebar to continue with this demonstration.",
            dest_rect=dest_rect,
            font_name='Arial', 
            font_size=36, 
            color=(25, 50, 255))

# Clear events before the text is drawn
response_handler.clear_events()


# Draw the text
running = True
while running:
    screen.fill((128, 128, 128))  # Gray background
    text.draw()  # Draw the text
    screen.flip()

    # Handle events
    response_handler.get_events()
    if response_handler.should_quit():
        running = False
        break

    # Example: Check if the spacebar was pressed
    elif response_handler.is_key_down('space'):
        print("Spacebar pressed!")
        running = False
        break


# Create a Text object
dest_rect = center_rect_on_point([0, 0, .85*screen.width, .2*screen.height], [center_x, center_y])
text = Text("Here is a shorter instruction. Press spacebar to quit.",
            dest_rect=dest_rect,
            font_name='Arial', 
            font_size=36, 
            color=(25, 50, 255))

# Clear events before the text is drawn
response_handler.clear_events()

# Draw the text
running = True
while running:
    screen.fill((128, 128, 128))  # Gray background
    text.draw()  # Draw the text
    screen.flip()

    # Handle events
    response_handler.get_events()
    if response_handler.should_quit():
        running = False
        break

    # Example: Check if the spacebar was pressed
    elif response_handler.is_key_down('space'):
        print("Spacebar pressed!")
        running = False
        break

# Analyze frame intervals after the loop ends
frame_intervals = np.array(frame_intervals)
average_interval = np.mean(frame_intervals) * 1000  # Convert to milliseconds
std_deviation = np.std(frame_intervals) * 1000      # Convert to milliseconds

print(f"Average frame interval: {average_interval:.4f} ms")
print(f"Standard deviation: {std_deviation:.4f} ms")

# one last flip
screen.flip()

# close the audio player
audio_player.close()

# close the screen
screen.close()



import numpy as np
import random
  
def fabriquer_grille_sin(nx, frequence, phase, angle):
    # fabriquer_grille_sin permet de fabriquer l'image carrée d'une grille sinusoïdale variant entre 0 et 1.
    # La fontion admet 4 variables:
    #     nx: spécifie la largeur et la hauter de l'image
    #     frequence: spécifie la fréquence de la grille sinusoïdale en cycles par largeur d'image
    #     phase: spécifie la phase de la grille sinusoïdale en radians
    #     angle: spécifie l'orientation de la grille sinusoïdale en radians
    x = np.linspace(0, 1, nx)
    xv, yv = np.meshgrid(x, x)
    rampe = (np.cos(angle) * xv + np.sin(angle) * yv)
    grille_sin = np.sin(frequence * 2 * np.pi * rampe + phase) / 2 + 0.5
    return grille_sin


def fabriquer_enveloppe_gaussienne(nx, ecart_type):
    # fabriquer_enveloppe_gaussienne permet de fabriquer l'image carrée d'une enveloppe gaussienne centrale variant entre 0 et 1.
    # La fontion admet 2 variables:
    #     nx: spécifie la largeur et la hauter de l'image
    #     ecart_type: spécifie l'écart-type de la gaussienne 2D en largeur d'image
    x = np.linspace(0, 1, nx)
    xv, yv = np.meshgrid(x, x)
    gaussienne = np.exp(-((xv - 0.5) ** 2 / ecart_type ** 2) - ((yv - 0.5) ** 2 / ecart_type ** 2))
    return gaussienne


def fabriquer_gabor(nx, frequence, phase, angle, ecart_type):
    # fabriquer_gabor permet de fabriquer une tache de gabor variant entre 0 et 1
    # La fontion admet 5 variables:
    #     nx: spécifie la largeur et la hauter de l'image
    #     frequence: spécifie la fréquence de la grille sinusoïdale en cycles par largeur d'image
    #     phase: spécifie la phase de la grille sinusoïdale en radians
    #     angle: spécifie l'orientation de la grille sinusoïdale en radians
    #     ecart_type: spécifie l'écart-type de la gaussienne 2D en largeur d'image
    gaussienne = fabriquer_enveloppe_gaussienne(nx, ecart_type)
    grille_sin = fabriquer_grille_sin(nx, frequence, phase, angle)
    gabor = gaussienne * (grille_sin - 0.5) + 0.5
    return gabor


def stretch(im):

  tim = (im - np.amin(im)) / (np.amax(im) - np.amin(im))
  return tim


def noisy_bit_dithering(im, depth = 256):
    # Implements the dithering algorithm presented in:
    # Allard, R., Faubert, J. (2008) The noisy-bit method for digital displays:
    # converting a 256 luminance resolution into a continuous resolution. Behavior 
    # Research Method, 40(3), 735-743.
    # It takes 2 arguments:
    #   im: is an image matrix in float64 that varies between 0 and 1, 
    #   depth: is the number of evenly separated luminance values at your disposal. 
    #     Default is 256 (1 byte).
    # It returns:
    #   tim: a matrix containg integer values between 1 and depth, indicating which 
    #     luminance value should be used for every pixel. 
    #
    # E.g.:
    #   tim = noisy_bit_dithering(im, depth = 256)
    #
    # This example assumes that all rgb values are linearly related to luminance 
    # values (e.g. on a Mac, put your LCD monitor gamma parameter to 1 in the Displays 
    # section of the System Preferences). If this is not the case, use a lookup table 
    # to transform the tim integer values into rgb values corresponding to evenly 
    # spaced luminance values.
    #
    # Frederic Gosselin, 27/09/2022
    # frederic.gosselin@umontreal.ca
        tim = im * (depth - 1.0)
        tim = np.uint8(np.fmax(np.fmin(np.around(tim + np.random.random(np.shape(im)) - 0.5), depth - 1.0), 0.0))
        return tim


def fabriquer_wiggles_sin(nx, frequence_min, frequence_max, frequence_radiale, phase_radiale, phase):
    # fabriquer_wiggles_sin permet de fabriquer l'image carrée de "wiggles", les stimuli inventés par Frances Wilkinson, variant entre 0 et 1.
    # La fontion admet 6 variables:
    #     nx: spécifie la largeur et la hauter de l'image
    #     frequence_radiale: spécifie le nombre de bosses et de creux par 2*pi radians
    #     phase_radiale: spécifie la phase angulaire des creux et des bossses en radians
    #     frequence_min: spécifie la fréquence des bosses concentriques par largeur d'image
    #     frequence_max: spécifie la fréquence des creux concentriques par largeur d'image
    #     phase: spécifie la phase des bosses et des creux concentriques en radians
    x = np.linspace(0, 1, nx)
    xv, yv = np.meshgrid(x, x)
    angles = np.arctan2((yv - 0.5), (xv - 0.5))
    modulation_freq = (frequence_max - frequence_min) * (np.sin(frequence_radiale * angles + phase_radiale) / 2 + 0.5) + frequence_min
    rayons = np.sqrt((xv - 0.5) ** 2 + (yv - 0.5) ** 2)
    wiggles_sin = np.sin(modulation_freq * 2 * np.pi * rayons + phase) / 2 + 0.5
    return wiggles_sin


def fabriquer_cercles_sin(nx, frequence, phase):
    # fabriquer_cercles_sin permet de fabriquer l'image carrée de cercles concentriques sinusoïdaux variant entre 0 et 1.
    # La fontion admet 3 variables:
    #     nx: spécifie la largeur et la hauter de l'image
    #     frequence: spécifie la fréquence des cercles sinusoïdaux en cycles par largeur d'image
    #     phase: spécifie la phase des cercles sinusoïdaux en radians
    x = np.linspace(0, 1, nx)
    xv, yv = np.meshgrid(x, x)
    rayons = np.sqrt((xv - 0.5) ** 2 + (yv - 0.5) ** 2)
    cercles_sin = np.sin(frequence * 2 * np.pi * rayons + phase) / 2 + 0.5;
    return cercles_sin


def fabriquer_secteurs_sin(nx, frequence, phase):
    # fabriquer_secteurs_sin permet de fabriquer l'image carrée de secteurs sinusoïdaux variant entre 0 et 1.
    # La fontion admet 3 variables:
    #     nx: spécifie la largeur et la hauter de l'image
    #     frequence: spécifie la fréquence des secteurs sinusoïdaux en cycles par 2*pi
    #     phase: spécifie la phase des secteurs sinusoïdaux en radians
    x = np.linspace(0, 1, nx)
    xv, yv = np.meshgrid(x, x)
    angles = np.arctan2((yv - 0.5), (xv - 0.5))
    secteurs_sin = np.sin(frequence * angles + phase) / 2 + 0.5
    return secteurs_sin


def fabriquer_grand_damier(une_case, M, N):
    petit_damier = fabriquer_petit_damier(une_case)
    grand_damier = np.zeros((2*M*une_case, 2*N*une_case))
    for xx in np.arange(M):
        for yy in np.arange(N):
            grand_damier[xx*2*une_case:(xx+1)*2*une_case:1, yy*2*une_case:(yy+1)*2*une_case:1] = petit_damier
    return grand_damier



def fabriquer_petit_damier(une_case):
    petit_damier = np.zeros((2*une_case, 2*une_case))
    petit_damier[0:une_case:1,0:une_case:1] = np.ones((une_case, une_case))
    petit_damier[une_case:2*une_case:1,une_case:2*une_case:1] = np.ones((une_case, une_case))
    return petit_damier


def location_bubbles(nb_bubbles=50, std_bubble=25, an_image=None, x_size=None, y_size=None, random_state=None):
    """
    Generates a noisy spatial mask (referred to as a 'bubbles mask') and applies it to an image, 
    as first done in Gosselin and Schyns (2001, Vision Research, Experiment 1). This function 
    is used both to generate stimuli for an experiment and the bubbles masks for a classification 
    image analysis afterward.

    Parameters:
    nb_bubbles : int, optional
        The expected number of bubbles (sampled areas). The exact number of bubbles varies from 
        trial to trial to ensure spatial independence of the bubbles. Default is 50.
    std_bubble : float, optional
        The standard deviation of each Gaussian filter (bubble) in pixels. Each bubble modulates 
        the contrast from 0 at the edges to 1 at the center. Default is 25.
    an_image : ndarray, optional
        The image to which the bubbles will be applied. This image can be grayscale (x_size x y_size) 
        or color (x_size x y_size x 3), and its values must range between 0 and 1. Must be provided if 
        x_size or y_size are not.
    x_size : int, optional
        The width of the image. Either x_size and y_size or an_image must be provided.
    y_size : int, optional
        The height of the image. Either x_size and y_size or an_image must be provided.
    random_state : random.RandomState object, optional
        A random state object to control reproducibility. Must be provided.

    Returns:
    tuple:
        If an_image is not provided:
            - the_noise: ndarray
                A (x_size + 2*std_bubble*5 by y_size + 2*std_bubble*5) binary array representing 
                the centers of the bubbles. It is larger than the image to ensure homogeneous sampling, 
                including at the image's borders.
        If an_image is provided:
            - the_noise: ndarray
                A (x_size + 2*std_bubble*n_zero by y_size + 2*std_bubble*n_zeros, with n_zero equal to 5) 
                binary array representing the centers of the bubbles. It is larger than the image to ensure 
                homogeneous sampling, including at the image's borders. 
            - stimulus: ndarray
                The resulting stimulus after applying the bubbles mask to the original image. Where the 
                bubbles do not sample the image, the stimulus appears mid-gray (value of 0.5).

    Raises:
    ValueError:
        If neither an image nor x_size and y_size are provided, or if random_state is not provided.

    Example:
        a_seed = 0
        np.random.seed(a_seed)
        x_size = 256
        y_size = 200
        an_image = np.random.random((y_size, x_size)) # grayscale image with values between 0 and 1        
        nb_bubbles = 100
        std_bubble = 15
        random_state = np.random.get_state()
        the_noise, stimulus = location_noise(nb_bubbles=nb_bubbles, std_bubble=std_bubble, an_image=an_image, x_size=None, y_size=None, random_state=random_state) # with an image
        #the_noise = location_noise(nb_bubbles=nb_bubbles, std_bubble=std_bubble, an_image=None, x_size=an_image.shape[0], y_size=an_image.shape[1], random_state=random_state) # with coordinates

    History:
        Written by Frederic Gosselin, October 17 2024
    """

    # Set the random state for reproducibility
    if random_state is not None:
        #random.setstate(random_state)
        np.random.set_state(random_state)
    else:
        raise ValueError("Must provide a pseudo-random generator state.")

    # Determine the size of the image (x_size, y_size) from the provided image if not explicitly provided
    if x_size is None and y_size is None and an_image is not None:
        y_size, x_size = an_image.shape[:2]  # Use the shape of the image

    # Raise an error if no valid size information is available
    if x_size is None or y_size is None:
        raise ValueError("Either x_size and y_size must be specified, or an_image must be provided.")
    
    # Generate probabilistic placement of bubbles over an extended area (to handle image borders)
    n_zero = 5  # Number of standard deviations to use for the bubble size
    max_half_size = round(std_bubble * n_zero)  # Maximum size (half-width) for the Gaussian bubble
    #temp_rand = np.asarray([[random.uniform(0, 1) for _ in range(x_size + 2*max_half_size)] for _ in range(y_size + 2*max_half_size)]) # Noise over extended area
    temp_rand = np.random.rand(y_size + 2*max_half_size, x_size + 2*max_half_size)
    the_noise = (temp_rand <= (nb_bubbles / ((x_size+2*max_half_size) * (y_size+2*max_half_size))))  # Binary mask

    # If no image is provided, return just the noise mask and the current random state
    if an_image is None:
        #return the_noise, random.getstate()
        return the_noise
    
    else:
        # Create a 2D Gaussian kernel to simulate bubble spread (blur effect)
        y, x = np.meshgrid(np.arange(-max_half_size, max_half_size + 1), np.arange(-max_half_size, max_half_size + 1))
        gauss = np.exp(-(x**2 / std_bubble**2 + y**2 / std_bubble**2))  # Gaussian equation
        gauss /= np.max(gauss)  # Normalize Gaussian kernel

        # Apply convolution using FFT (efficient for large arrays)
        f_the_noise = np.fft.fft2(the_noise.astype(float), s=(y_size + 4*max_half_size, x_size + 4*max_half_size))  # FFT of noise
        f_padded_gauss = np.fft.fft2(gauss, s=(y_size + 4*max_half_size, x_size + 4*max_half_size))  # FFT of Gaussian
        temp_plane = np.fft.ifft2(f_the_noise * f_padded_gauss).real  # Inverse FFT to apply Gaussian blur

        # Extract the valid region of the resulting plane after convolution
        win_plane = np.minimum(temp_plane[max_half_size:y_size + max_half_size, max_half_size:x_size + max_half_size], 1)

        # Handle color images by applying the noise plane to each channel
        if len(an_image.shape) == 3:
            win_plane = np.stack((win_plane,) * 3, axis=-1)  # Stack the plane for RGB channels

        # Combine the noise mask with the original image to generate the final stimulus
        stimulus = win_plane / 2 * (an_image - 0.5) + 0.5  # Blend noise mask with image

        # Return the noise mask, the final stimulus, and the current random state
        return the_noise, stimulus



# setup.py

from setuptools import setup, find_packages

requires = []
with open('requirements.txt') as reqfile:
    requires = reqfile.read().splitlines()


setup(
    name='tachypy',
    version='0.1.2',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    install_requires=requires,
    python_requires='>=3.6',
    author='Ian Charest and Frederic Gosselin',
    author_email='charest.ian@gmail.com',
    description='A package for OpenGL drawing using Pygame.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/CharestLab/tachypy',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Operating System :: OS Independent',
    ],
)



import sounddevice as sd
import numpy as np
import threading
import time

class Audio:
    def __init__(self, sample_rate=44100, channels=1):
        self.sample_rate = sample_rate
        self.channels = channels
        self.is_playing = False

    def play(self, data, when=0):
        """
        Plays a NumPy array as audio at a specified time.
        :param data: NumPy array of audio data.
        :param when: Absolute time in seconds (monotonic) when playback should start.
        """
        if not isinstance(data, np.ndarray):
            raise ValueError("Data must be a NumPy array.")

        # Ensure data has the correct shape
        if data.ndim == 1:
            if self.channels > 1:
                # Duplicate mono data for multiple channels
                data = np.tile(data[:, np.newaxis], (1, self.channels))
        elif data.ndim == 2:
            if data.shape[1] != self.channels:
                raise ValueError(f"Data must have {self.channels} channels.")
        else:
            raise ValueError("Data must be a 1D or 2D NumPy array.")

        # Ensure data is in float32 format
        if data.dtype != np.float32:
            data = data.astype(np.float32)

        # Calculate delay until 'when' time
        current_time = time.monotonic_ns()
        delay = (when * 1e9) - current_time # convert delay from secs to nanosecs

        if delay < 0:
            print("Warning: 'when' time is in the past. Playing immediately.")
            delay = 0

        # Start playback in a separate thread to avoid blocking
        threading.Thread(target=self._playback_thread, args=(data, delay), daemon=True).start()

    def _playback_thread(self, data, delay):
        if delay > 0:
            # High-precision wait until the scheduled time
            self._precise_delay(delay)

        self.is_playing = True
        sd.play(data, samplerate=self.sample_rate)
        sd.wait()  # Wait until playback is finished
        self.is_playing = False

    def _precise_delay(self, delay):
        """Wait for the specified delay with high precision."""
        end_time = time.monotonic_ns() + delay
        while True:
            remaining = end_time - time.monotonic_ns()
            if remaining <= 0:
                break
            elif remaining > 0.01:
                # Sleep for a shorter duration to improve timing accuracy
                time.sleep(0.005)
            else:
                # Busy-wait for higher precision
                pass

    def stop(self):
        sd.stop()
        self.is_playing = False

    def close(self):
        self.stop()


# responses.py
import numpy as np
import pygame
from OpenGL.GL import *
from OpenGL.GLU import *
from tachypy.shapes import Line, center_rect_on_point  # Assuming Line class is in your visuals module
from tachypy.text import Text     # Assuming you have a Text class
import time
# responses.py


class ResponseHandler:
    """
    A class to handle user input events in a Pygame application.

    This class processes keyboard and mouse events, maintains the current state
    of keys and mouse buttons, and provides methods to access this information.
    It also handles quitting the application when needed.

    Attributes:
        should_exit (bool): Flag indicating whether the application should quit.
        start_time (int): Start time in nanoseconds for reaction time measurement.
        key_presses (list): List of key press and release events.
        mouse_clicks (list): List of mouse click events.
        key_down_events (set): Set of keys currently pressed down.
        key_up_events (set): Set of keys released in the current frame.
        mouse_position (tuple): Current position of the mouse cursor.
        mouse_buttons (tuple): Current state of mouse buttons (pressed or not).
        keys_to_listen (list): List of key names to listen for. If None, listens to all keys.
        events (list): List of Pygame events processed in the current frame.
    """

    def __init__(self, keys_to_listen=None):
        """
        Initialize the ResponseHandler.

        Parameters:
            keys_to_listen (list): List of key names to listen for (e.g., ['a', 's', 'escape']).
                                   If None, listens to all keys.
        """
        # Flag indicating whether the application should quit
        self.should_exit = False
        # Start time for reaction time measurement
        self.start_time = time.monotonic_ns()

        # Lists to store input events
        self.key_presses = []    # List of key press and release events
        self.mouse_clicks = []   # List of mouse click events

        # Sets to track current key states
        self.key_down_events = set()  # Set of keys currently pressed down
        self.key_up_events = set()    # Set of keys released in the current frame

        # Mouse state
        self.mouse_position = None                   # Current mouse position (x, y)
        self.mouse_buttons = [False, False, False]   # Mouse buttons states: [Left, Middle, Right]

        # List of key names to listen for
        self.keys_to_listen = keys_to_listen

        # List to store Pygame events processed in the current frame
        self.events = []

    def reset_timer(self):
        """
        Reset the start time for reaction time measurements.
        """
        self.start_time = time.monotonic_ns()

    def get_events(self):
        """
        Retrieve and process Pygame events, updating internal state.

        Stores the events in self.events for access by other components.
        """
        # Clear key event sets
        self.key_down_events.clear()
        self.key_up_events.clear()

        # Get events from Pygame event queue
        self.events = pygame.event.get()

        for event in self.events:
            # Calculate timestamp relative to start_time, in seconds
            timestamp = (time.monotonic_ns() - self.start_time) / 1e9

            if event.type == pygame.QUIT:
                # Set flag to exit the application
                self.should_exit = True

            elif event.type == pygame.KEYDOWN:
                # Key pressed down
                key_name = pygame.key.name(event.key)
                if self.keys_to_listen is None or key_name in self.keys_to_listen:
                    # Store key press event
                    self.key_presses.append({
                        'time': timestamp,
                        'type': 'keydown',
                        'key': key_name
                    })
                    # Add to set of currently pressed keys
                    self.key_down_events.add(key_name)
                    # Check for escape key to quit application
                    if key_name == 'escape':
                        self.should_exit = True

            elif event.type == pygame.KEYUP:
                # Key released
                key_name = pygame.key.name(event.key)
                if self.keys_to_listen is None or key_name in self.keys_to_listen:
                    # Store key release event
                    self.key_presses.append({
                        'time': timestamp,
                        'type': 'keyup',
                        'key': key_name
                    })
                    # Add to set of keys released this frame
                    self.key_up_events.add(key_name)

            elif event.type == pygame.MOUSEBUTTONDOWN:
                # Mouse button pressed
                self.mouse_clicks.append({
                    'time': timestamp,
                    'type': 'mousedown',
                    'button': event.button - 1,  # Adjust to 0-based index
                    'pos': event.pos
                })

            elif event.type == pygame.MOUSEBUTTONUP:
                # Mouse button released
                self.mouse_clicks.append({
                    'time': timestamp,
                    'type': 'mouseup',
                    'button': event.button - 1,  # Adjust to 0-based index
                    'pos': event.pos
                })

        # Update mouse position and button states
        self.mouse_position = pygame.mouse.get_pos()
        self.mouse_buttons = pygame.mouse.get_pressed()

    def should_quit(self):
        """
        Check if the application should quit.

        Returns:
            bool: True if the application should quit, False otherwise.
        """
        return self.should_exit

    def get_key_presses(self):
        """
        Get the list of key press and release events.

        Returns:
            list: A list of dictionaries with keys:
                - 'time': Timestamp of the event relative to start_time (float).
                - 'type': 'keydown' or 'keyup'.
                - 'key': Name of the key.
        """
        return self.key_presses

    def is_key_down(self, key_name):
        """
        Check if a specific key is currently pressed down.

        Parameters:
            key_name (str): Name of the key to check.

        Returns:
            bool: True if the key is currently pressed, False otherwise.
        """
        return key_name in self.key_down_events

    def get_mouse_position(self):
        """
        Get the current mouse position.

        Returns:
            tuple: (x, y) coordinates of the mouse cursor.
        """
        return self.mouse_position

    def is_mouse_button_pressed(self, button):
        """
        Check if a specific mouse button is currently pressed.

        Parameters:
            button (int): 0 for left button, 1 for middle button, 2 for right button.

        Returns:
            bool: True if the button is currently pressed, False otherwise.
        """
        return self.mouse_buttons[button]

    def get_mouse_clicks(self):
        """
        Get the list of mouse click events.

        Returns:
            list: A list of dictionaries with keys:
                - 'time': Timestamp of the event relative to start_time (float).
                - 'type': 'mousedown' or 'mouseup'.
                - 'button': 0 for left button, 1 for middle button, 2 for right button.
                - 'pos': (x, y) position of the mouse at the time of the event.
        """
        return self.mouse_clicks

    def set_position(self, x, y):
        """
        Set the mouse cursor position.

        Parameters:
            x (int): x-coordinate to move the cursor to.
            y (int): y-coordinate to move the cursor to.
        """
        pygame.mouse.set_pos((x, y))
        # Update internal mouse position
        self.mouse_position = (x, y)

    def clear_events(self):
        """
        Clear the lists of key presses and mouse clicks.

        This can be used to reset the stored events after they've been processed.
        """
        self.key_presses.clear()
        self.mouse_clicks.clear()
        self.key_down_events.clear()
        self.key_up_events.clear()
        self.should_exit = False
        # Get events from Pygame event queue
        self.events = pygame.event.get()


class Scrollbar:
    def __init__(self,
                 screen_width,
                 screen_height,
                 position_y=200,  # Since origin is at top-left, positive y is downward
                 half_bar_length=400,
                 bar_thickness=4,
                 bar_color=(0, 0, 0),
                 half_mark_height=5,
                 mark_thickness=3,
                 mark_color=(0, 0, 0),
                 num_marks=10,
                 half_end_height=20,
                 end_thickness=4,
                 end_color=(0, 0, 0),
                 text_left='0',
                 text_right='100',
                 font_size=24,
                 font_name='Helvetica',
                 text_color=(0, 0, 0),
                 text_offset=12):
        """
        Initialize the Scrollbar.

        Parameters:
            screen_width: Width of the screen.
            screen_height: Height of the screen.
            ... (other parameters remain the same) ...
        """
        self.screen_width = screen_width
        self.screen_height = screen_height
        self.position_y = position_y
        self.half_bar_length = half_bar_length
        self.bar_thickness = bar_thickness
        self.bar_color = bar_color
        self.half_mark_height = half_mark_height
        self.mark_thickness = mark_thickness
        self.mark_color = mark_color
        self.num_marks = num_marks
        self.half_end_height = half_end_height
        self.end_thickness = end_thickness
        self.end_color = end_color
        self.text_left = text_left
        self.text_right = text_right
        self.text_size = font_size
        self.font_name = font_name
        self.text_color = text_color
        self.text_offset = text_offset

        # Center x-coordinate
        self.center_x = self.screen_width / 2

        # Create the main bar
        self.bar = Line(
            start_point=(self.center_x - self.half_bar_length, self.position_y),
            end_point=(self.center_x + self.half_bar_length, self.position_y),
            thickness=self.bar_thickness,
            color=self.bar_color
        )

        # Create the marks
        self.marks = []
        for x in np.linspace(self.center_x - self.half_bar_length,
                             self.center_x + self.half_bar_length,
                             self.num_marks):
            mark = Line(
                start_point=(x, self.position_y - self.half_mark_height),
                end_point=(x, self.position_y + self.half_mark_height),
                thickness=self.mark_thickness,
                color=self.mark_color
            )
            self.marks.append(mark)

        # Create the ends
        self.left_end = Line(
            start_point=(self.center_x - self.half_bar_length, self.position_y - self.half_end_height),
            end_point=(self.center_x - self.half_bar_length, self.position_y + self.half_end_height),
            thickness=self.end_thickness,
            color=self.end_color
        )
        self.right_end = Line(
            start_point=(self.center_x + self.half_bar_length, self.position_y - self.half_end_height),
            end_point=(self.center_x + self.half_bar_length, self.position_y + self.half_end_height),
            thickness=self.end_thickness,
            color=self.end_color
        )

        left_text_pos = center_rect_on_point(
            [0, 0, 500, 500], 
            [self.center_x - self.half_bar_length, self.position_y + self.half_end_height + self.text_offset]
        )

        right_text_pos = center_rect_on_point(
            [0, 0, 500, 500], 
            [self.center_x + self.half_bar_length, self.position_y + self.half_end_height + self.text_offset]
        )

        # Create the text labels
        self.text_left_label = Text(
            text=self.text_left,
            font_name=self.font_name,
            font_size=self.text_size,
            color=self.text_color,
            dest_rect = left_text_pos
        )
        self.text_right_label = Text(
            text=self.text_right,
            font_name=self.font_name,
            font_size=self.text_size,
            color=self.text_color,
            dest_rect = right_text_pos
        )

        # Mobile part (draggable line)
        self.half_mobile_line_height = 12    # Half the height of the mobile line
        self.mobile_line_thickness = 6       # Thickness of the mobile line
        self.mobile_line_color = (255, 0, 0) # Color of the mobile line (red)
        self.mobile_line_x = self.center_x   # Start at the center
        self.mobile_line = Line(
            start_point=(self.mobile_line_x, self.position_y - self.half_mobile_line_height),
            end_point=(self.mobile_line_x, self.position_y + self.half_mobile_line_height),
            thickness=self.mobile_line_thickness,
            color=self.mobile_line_color
        )

    def draw(self):
        # Draw the main bar
        self.bar.draw()
        # Draw the marks
        for mark in self.marks:
            mark.draw()
        # Draw the ends
        self.left_end.draw()
        self.right_end.draw()
        # Draw the text labels
        self.text_left_label.draw()
        self.text_right_label.draw()
        # Draw the mobile line
        self.mobile_line.draw()

    def handle_mouse(self, mouse_x, mouse_y):
        """
        Update the position of the mobile line based on mouse input.

        Parameters:
            mouse_x: x-coordinate of the mouse.
            mouse_y: y-coordinate of the mouse.
        """
        # Check if mouse_y is near the bar's position_y (within some tolerance)
        if abs(mouse_y - self.position_y) <= self.half_end_height * 2:
            # Clamp mouse_x within the bar's range
            min_x = self.center_x - self.half_bar_length
            max_x = self.center_x + self.half_bar_length
            new_x = np.clip(mouse_x, min_x, max_x)
            self.mobile_line_x = new_x
            # Update the mobile line's position
            self.mobile_line.set_start_point((self.mobile_line_x, self.position_y - self.half_mobile_line_height))
            self.mobile_line.set_end_point((self.mobile_line_x, self.position_y + self.half_mobile_line_height))
 

    def get_value(self):
        """
        Get the value corresponding to the position of the mobile line.

        Returns:
            A float value between 0 and 100.
        """
        min_x = self.center_x - self.half_bar_length
        max_x = self.center_x + self.half_bar_length
        value = ((self.mobile_line_x - min_x) / (max_x - min_x)) * 100
        return value
